@@ -1,13 +1,19 @@
 /*!
     # fltk-egui
 
-    An FLTK backend for Egui using a GlWindow. The code is largely based on https://github.com/ArjunNair/egui_sdl2_gl modified for fltk-rs.
+    An FLTK backend for Egui using a GlWindow.
 
     ## Usage
     Add to your Cargo.toml:
     ```toml
     [dependencies]
-    fltk-egui = "0.5" # targets egui 0.16
+    fltk-egui = "0.6" # targets egui 0.17
+    ```
+
+    On wayland desktop enable feature:
+    ```toml
+    [dependencies]
+    fltk-egui = { version = "0.6",  features = ["wayland"] }
     ```
 
     The basic premise is that egui is an immediate mode gui, while FLTK is retained.
@@ -31,28 +37,38 @@ use std::time::Instant;
 
 // Re-export dependencies.
 pub use egui;
-use egui::{pos2, CursorIcon, Event, Key, Modifiers, Pos2, RawInput, Vec2};
+use egui::{pos2, vec2, CursorIcon, Event, Key, Modifiers, Pos2, RawInput, Rect, Vec2};
+use egui_glow::Painter;
+pub use egui_glow::{glow, painter};
+pub use egui_image::RetainedEguiImage;
 pub use fltk;
 use fltk::{
     app, enums,
-    prelude::{FltkError, GroupExt, ImageExt, InputExt, WidgetExt, WindowExt},
+    prelude::{FltkError, ImageExt, WidgetExt, WindowExt},
     window::GlWindow,
 };
-pub use gl;
-mod painter;
-pub use painter::Painter;
+use glow::HasContext;
+mod clipboard;
+mod egui_image;
+use clipboard::Clipboard;
 
 /// Construct the backend.
-/// Requires the DpiScaling, which can be Default or Custom(f32)
-pub fn with_fltk(win: &mut GlWindow, scale: DpiScaling) -> (Painter, EguiInputState) {
-    let scale = match scale {
-        DpiScaling::Default => win.pixels_per_unit(),
-        DpiScaling::Custom(custom) => custom,
+pub fn with_fltk(win: &mut GlWindow) -> (glow::Context, Painter, EguiState) {
+    app::set_screen_scale(win.screen_num(), 1.);
+    app::keyboard_screen_scaling(false);
+    let gl = unsafe { glow::Context::from_loader_function(|s| win.get_proc_address(s) as _) };
+
+    unsafe {
+        // to fix black textured.
+        gl.enable(glow::FRAMEBUFFER_SRGB);
+        // to enable MULTISAMPLE.
+        gl.enable(glow::MULTISAMPLE)
     };
-    let inp = fltk::input::Input::default();
-    let painter = Painter::new(win, scale);
-    win.add(&inp);
-    EguiInputState::new(painter)
+
+    let painter = Painter::new(&gl, None, "")
+        .unwrap_or_else(|error| panic!("some OpenGL error occurred {}\n", error));
+    let max_texture_side = painter.max_texture_side();
+    (gl, painter, EguiState::new(&win, max_texture_side))
 }
 
 /// Frame time for FPS.
@@ -60,12 +76,11 @@ pub fn get_frame_time(start_time: Instant) -> f32 {
     (Instant::now() - start_time).as_secs_f64() as f32
 }
 
-/// The scaling factors of the app
-pub enum DpiScaling {
-    /// Default DPI Scale by fltk, usually 1.0
-    Default,
-    /// Custome DPI scaling, e.g: 1.5, 2.0 and so fort.
-    Custom(f32),
+/// Casting slice to another type of slice
+pub fn cast_slice<T, D>(s: &[T]) -> &[D] {
+    unsafe {
+        std::slice::from_raw_parts(s.as_ptr() as *const D, s.len() * std::mem::size_of::<T>())
+    }
 }
 
 /// The default cursor
@@ -89,30 +104,62 @@ impl Default for FusedCursor {
 }
 
 /// Shuttles FLTK's input and events to Egui
-pub struct EguiInputState {
+pub struct EguiState {
+    pub canvas_size: [u32; 2],
+    pub clipboard: Clipboard,
     pub fuse_cursor: FusedCursor,
-    pub pointer_pos: Pos2,
     pub input: RawInput,
-    pub modifiers: Modifiers,
+    _pixels_per_point: f32,
+    pub pointer_pos: Pos2,
+    /// default value is 12.0
+    pub scroll_factor: f32,
+    /// default value is 8.0
+    pub zoom_factor: f32,
     /// Internal use case for fn window_resized()
-    pub _window_resized: bool,
+    _window_resized: bool,
+    pub max_texture_side: usize,
 }
 
-impl EguiInputState {
+impl EguiState {
     /// Construct a new state
-    pub fn new(painter: Painter) -> (Painter, EguiInputState) {
-        let _self = EguiInputState {
+    pub fn new(win: &GlWindow, max_texture_side: usize) -> EguiState {
+        let ppu = win.pixels_per_unit();
+        let (width, height) = (win.width(), win.height());
+        let rect = vec2(width as f32, height as f32) / ppu;
+        let screen_rect = Rect::from_min_size(Pos2::new(0f32, 0f32), rect);
+        EguiState {
+            canvas_size: [width as u32, height as u32],
+            clipboard: Clipboard::default(),
             fuse_cursor: FusedCursor::new(),
-            pointer_pos: Pos2::new(0f32, 0f32),
             input: egui::RawInput {
-                screen_rect: Some(painter.screen_rect),
-                pixels_per_point: Some(painter.pixels_per_point),
+                screen_rect: Some(screen_rect),
+                pixels_per_point: Some(ppu),
+                max_texture_side: Some(max_texture_side),
                 ..Default::default()
             },
-            modifiers: Modifiers::default(),
+            max_texture_side,
+            _pixels_per_point: ppu,
+            pointer_pos: Pos2::new(0f32, 0f32),
+            scroll_factor: 12.0,
+            zoom_factor: 8.0,
             _window_resized: false,
-        };
-        (painter, _self)
+        }
+    }
+
+    /// Don't use state.input.take() use this fn instead (to avoid pixels per point miscalculation).
+    pub fn take_input(&mut self) -> egui::RawInput {
+        self.input.max_texture_side = Some(self.max_texture_side);
+        let pixels_per_point = self.input.pixels_per_point;
+        let take = self.input.take();
+        self.input.pixels_per_point = pixels_per_point;
+        if let Some(ppp) = pixels_per_point {
+            self._pixels_per_point = ppp;
+        }
+        take
+    }
+
+    pub fn pixels_per_point(&self) -> f32 {
+        self._pixels_per_point
     }
 
     /// Check if current window being resized.
@@ -123,34 +170,53 @@ impl EguiInputState {
     }
 
     /// Conveniece method bundling the necessary components for input/event handling
-    pub fn fuse_input(&mut self, win: &mut GlWindow, event: enums::Event, painter: &mut Painter) {
-        input_to_egui(win, event, self, painter);
+    pub fn fuse_input(&mut self, win: &mut GlWindow, event: enums::Event) {
+        input_to_egui(win, event, self);
     }
 
     /// Convenience method for outputting what egui emits each frame
-    pub fn fuse_output(&mut self, win: &mut GlWindow, egui_output: &egui::Output) {
+    pub fn fuse_output(&mut self, win: &mut GlWindow, egui_output: egui::PlatformOutput) {
+        if !egui_output.copied_text.is_empty() {
+            self.clipboard.set(egui_output.copied_text);
+        }
+        translate_cursor(win, &mut self.fuse_cursor, egui_output.cursor_icon);
+    }
+
+    /// Convenience method for outputting what egui emits each frame (borrow PlatformOutput)
+    pub fn fuse_output_borrow(&mut self, win: &mut GlWindow, egui_output: &egui::PlatformOutput) {
         if !egui_output.copied_text.is_empty() {
             app::copy(&egui_output.copied_text);
         }
         translate_cursor(win, &mut self.fuse_cursor, egui_output.cursor_icon);
     }
+
+    /// Set visual scale, e.g: 0.8, 1.5, 2.0 .etc (default is 1.0)
+    pub fn set_visual_scale(&mut self, size: f32) {
+        // have to be setted the pixels_per_point of both the inner (input) and the state.
+        self.input.pixels_per_point = Some(size);
+        self._pixels_per_point = size;
+
+        // resize rect with canvas.
+        let canvas_size = self.canvas_size;
+        let rect = vec2(canvas_size[0] as f32, canvas_size[1] as f32) / size;
+        self.input.screen_rect = Some(Rect::from_min_size(Default::default(), rect));
+    }
 }
 
 /// Handles input/events from FLTK
 pub fn input_to_egui(
     win: &mut GlWindow,
     event: enums::Event,
-    state: &mut EguiInputState,
-    painter: &mut Painter,
+    state: &mut EguiState,
+    // painter: &mut Painter,
 ) {
-    let (x, y) = app::event_coords();
-    let pixels_per_point = painter.pixels_per_point;
     match event {
         enums::Event::Resize => {
-            painter.update_screen_rect((win.width(), win.height()));
-            state.input.screen_rect = Some(painter.screen_rect);
+            state.canvas_size = [win.width() as u32, win.height() as u32];
+            state.set_visual_scale(state.pixels_per_point());
             state._window_resized = true;
         }
+
         //MouseButonLeft pressed is the only one needed by egui
         enums::Event::Push => {
             let mouse_btn = match app::event_mouse_button() {
@@ -164,7 +230,7 @@ pub fn input_to_egui(
                     pos: state.pointer_pos,
                     button: pressed,
                     pressed: true,
-                    modifiers: state.modifiers,
+                    modifiers: state.input.modifiers,
                 })
             }
         }
@@ -183,13 +249,15 @@ pub fn input_to_egui(
                     pos: state.pointer_pos,
                     button: released,
                     pressed: false,
-                    modifiers: state.modifiers,
+                    modifiers: state.input.modifiers,
                 })
             }
         }
 
         enums::Event::Move | enums::Event::Drag => {
-            state.pointer_pos = pos2(x as f32 / pixels_per_point, y as f32 / pixels_per_point);
+            let ppp = state.pixels_per_point();
+            let (x, y) = app::event_coords();
+            state.pointer_pos = pos2(x as f32 / ppp, y as f32 / ppp);
             state
                 .input
                 .events
@@ -199,7 +267,7 @@ pub fn input_to_egui(
         enums::Event::KeyUp => {
             if let Some(key) = translate_virtual_key_code(app::event_key()) {
                 let keymod = app::event_state();
-                state.modifiers = Modifiers {
+                state.input.modifiers = Modifiers {
                     alt: (keymod & enums::EventState::Alt == enums::EventState::Alt),
                     ctrl: (keymod & enums::EventState::Ctrl == enums::EventState::Ctrl),
                     shift: (keymod & enums::EventState::Shift == enums::EventState::Shift),
@@ -208,10 +276,10 @@ pub fn input_to_egui(
                     //TOD: Test on both windows and mac
                     command: (keymod & enums::EventState::Command == enums::EventState::Command),
                 };
-                if state.modifiers.command && key == Key::V {
-                    let inp: fltk::input::Input = unsafe { win.child(0).unwrap().into_widget() };
-                    app::paste(&inp);
-                    state.input.events.push(Event::Text(inp.value()));
+                if state.input.modifiers.command && key == Key::V {
+                    if let Some(value) = state.clipboard.get() {
+                        state.input.events.push(egui::Event::Text(value));
+                    }
                 }
             }
         }
@@ -227,7 +295,7 @@ pub fn input_to_egui(
             }
             if let Some(key) = translate_virtual_key_code(app::event_key()) {
                 let keymod = app::event_state();
-                state.modifiers = Modifiers {
+                state.input.modifiers = Modifiers {
                     alt: (keymod & enums::EventState::Alt == enums::EventState::Alt),
                     ctrl: (keymod & enums::EventState::Ctrl == enums::EventState::Ctrl),
                     shift: (keymod & enums::EventState::Shift == enums::EventState::Shift),
@@ -240,20 +308,20 @@ pub fn input_to_egui(
                 state.input.events.push(Event::Key {
                     key,
                     pressed: true,
-                    modifiers: state.modifiers,
+                    modifiers: state.input.modifiers,
                 });
 
-                if state.modifiers.command && key == Key::C {
+                if state.input.modifiers.command && key == Key::C {
                     // println!("copy event");
                     state.input.events.push(Event::Copy)
-                } else if state.modifiers.command && key == Key::X {
+                } else if state.input.modifiers.command && key == Key::X {
                     // println!("cut event");
                     state.input.events.push(Event::Cut)
                 } else {
                     state.input.events.push(Event::Key {
                         key,
                         pressed: false,
-                        modifiers: state.modifiers,
+                        modifiers: state.input.modifiers,
                     })
                 }
             }
@@ -261,23 +329,35 @@ pub fn input_to_egui(
 
         enums::Event::MouseWheel => {
             if app::is_event_ctrl() {
-                let zoom_factor = 1.2;
+                let zoom_factor = state.zoom_factor;
                 match app::event_dy() {
                     app::MouseWheel::Up => {
-                        state.input.events.push(Event::Zoom(zoom_factor * -1.0));
+                        let delta = egui::vec2(1., -1.) * zoom_factor;
+
+                        // Treat as zoom in:
+                        state
+                            .input
+                            .events
+                            .push(Event::Zoom((delta.y / 200.0).exp()));
                     }
                     app::MouseWheel::Down => {
-                        state.input.events.push(Event::Zoom(zoom_factor));
+                        let delta = egui::vec2(-1., 1.) * zoom_factor;
+
+                        // Treat as zoom out:
+                        state
+                            .input
+                            .events
+                            .push(Event::Zoom((delta.y / 200.0).exp()));
                     }
                     _ => (),
                 }
             } else {
-                let scroll_factor = 15.0;
+                let scroll_factor = state.scroll_factor;
                 match app::event_dy() {
                     app::MouseWheel::Up => {
                         state.input.events.push(Event::Scroll(Vec2 {
-                            x: scroll_factor,
-                            y: 0.,
+                            x: 0.,
+                            y: -scroll_factor,
                         }));
                     }
                     app::MouseWheel::Down => {
@@ -400,65 +480,103 @@ pub trait EguiImageConvertible<I>
 where
     I: ImageExt,
 {
-    fn to_egui_image(
-        self,
-        painter: &mut Painter,
-        new_size: (u32, u32),
-        filtering: bool,
-    ) -> Result<(egui::Image, egui::TextureId), FltkError>;
+    fn egui_image(self, debug_name: &str) -> Result<RetainedEguiImage, FltkError>;
 }
 
 impl<I> EguiImageConvertible<I> for I
 where
     I: ImageExt,
 {
-    /// Return (egui::Image, egui::TextureId)
-    fn to_egui_image(
-        self,
-        painter: &mut Painter,
-        new_size: (u32, u32),
-        filtering: bool,
-    ) -> Result<(egui::Image, egui::TextureId), FltkError> {
-        let size = (self.data_w() as usize, self.data_h() as usize);
-        let texture_id = painter.new_user_texture_rgba8(
+    /// Return (egui_extras::RetainedImage)
+    fn egui_image(self, debug_name: &str) -> Result<RetainedEguiImage, FltkError> {
+        let size = [self.data_w() as usize, self.data_h() as usize];
+        let color_image = egui::ColorImage::from_rgba_unmultiplied(
             size,
-            self.to_rgb()?
+            &self
+                .to_rgb()?
                 .convert(enums::ColorDepth::Rgba8)?
                 .to_rgb_data(),
-            filtering,
         );
-        let image = egui::Image::new(texture_id, egui::vec2(new_size.0 as _, new_size.1 as _));
-        Ok((image, texture_id))
+
+        Ok(RetainedEguiImage::from_color_image(debug_name, color_image))
     }
 }
 
 pub trait EguiSvgConvertible {
-    fn to_egui_image(
-        self,
-        painter: &mut Painter,
-        new_size: (u32, u32),
-        filtering: bool,
-    ) -> Result<(egui::Image, egui::TextureId), FltkError>;
+    fn egui_svg_image(self, debug_name: &str) -> Result<RetainedEguiImage, FltkError>;
 }
 
 impl EguiSvgConvertible for fltk::image::SvgImage {
-    /// Return (egui::Image, egui::TextureId)
-    fn to_egui_image(
-        mut self,
-        painter: &mut Painter,
-        new_size: (u32, u32),
-        filtering: bool,
-    ) -> Result<(egui::Image, egui::TextureId), FltkError> {
+    /// Return (egui_extras::RetainedEguiImage)
+    fn egui_svg_image(mut self, debug_name: &str) -> Result<RetainedEguiImage, FltkError> {
         self.normalize();
-        let size = (self.data_w() as usize, self.data_h() as usize);
-        let texture_id = painter.new_user_texture_rgba8(
+        let size = [self.data_w() as usize, self.data_h() as usize];
+        let color_image = egui::ColorImage::from_rgba_unmultiplied(
             size,
-            self.to_rgb()?
+            &self
+                .to_rgb()?
                 .convert(enums::ColorDepth::Rgba8)?
                 .to_rgb_data(),
-            filtering,
         );
-        let image = egui::Image::new(texture_id, egui::vec2(new_size.0 as _, new_size.1 as _));
-        Ok((image, texture_id))
+
+        Ok(RetainedEguiImage::from_color_image(debug_name, color_image))
     }
 }
+
+/// egui::TextureHandle from Vec egui::Color32
+pub fn tex_handle_from_vec_color32(
+    ctx: &egui::Context,
+    debug_name: &str,
+    vec: Vec<egui::Color32>,
+    size: [usize; 2],
+) -> egui::TextureHandle {
+    let mut pixels: Vec<u8> = Vec::with_capacity(vec.len() * 4);
+    vec.into_iter().for_each(|x| {
+        pixels.push(x[0]);
+        pixels.push(x[1]);
+        pixels.push(x[2]);
+        pixels.push(x[3]);
+    });
+    let color_image = egui::ColorImage::from_rgba_unmultiplied(size, pixels.as_slice());
+    ctx.load_texture(debug_name, color_image)
+}
+
+/// egui::TextureHandle from slice of egui::Color32
+pub fn tex_handle_from_color32_slice(
+    ctx: &egui::Context,
+    debug_name: &str,
+    slice: &[egui::Color32],
+    size: [usize; 2],
+) -> egui::TextureHandle {
+    let mut pixels: Vec<u8> = Vec::with_capacity(slice.len() * 4);
+    slice.into_iter().for_each(|x| {
+        pixels.push(x[0]);
+        pixels.push(x[1]);
+        pixels.push(x[2]);
+        pixels.push(x[3]);
+    });
+    let color_image = egui::ColorImage::from_rgba_unmultiplied(size, pixels.as_slice());
+    ctx.load_texture(debug_name, color_image)
+}
+
+/// egui::TextureHandle from slice of u8
+pub fn tex_handle_from_u8_slice(
+    ctx: &egui::Context,
+    debug_name: &str,
+    slice: &[u8],
+    size: [usize; 2],
+) -> egui::TextureHandle {
+    let color_image = egui::ColorImage::from_rgba_unmultiplied(size, slice);
+    ctx.load_texture(debug_name, color_image)
+}
+
+/// egui::TextureHandle from Vec u8
+pub fn tex_handle_from_vec_u8(
+    ctx: &egui::Context,
+    debug_name: &str,
+    vec: Vec<u8>,
+    size: [usize; 2],
+) -> egui::TextureHandle {
+    let color_image = egui::ColorImage::from_rgba_unmultiplied(size, vec.as_slice());
+    ctx.load_texture(debug_name, color_image)
+}
